import os
import platform
import subprocess
import sys
import threading
from typing import Optional

from hbox.logger import get_logger

log = get_logger(__name__)


def reader(pipe, func):
    for line in iter(pipe.readline, b''):
        func(line.decode())


def execute_in_shell(command, can_be_interactive):
    stdin_pipe = subprocess.PIPE if can_be_interactive else None
    process = subprocess.Popen(' '.join(command), shell=True, stdin=stdin_pipe,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return process


def execute_in_subprocess(command, can_be_interactive):
    stdin_pipe = subprocess.PIPE if can_be_interactive else None
    process = subprocess.Popen(command, stdin=stdin_pipe, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    return process


def execute_command(command, can_be_interactive: bool = False) -> Optional[int]:
    command_execution_func = execute_in_shell if platform.system().lower() == "linux" else execute_in_subprocess
    process = None
    stdin_data = None
    return_code = -1

    if can_be_interactive:
        if not sys.stdin.isatty():
            stdin_data = sys.stdin.read()
        if stdin_data:
            command.insert(2, "-i")

    try:
        log.debug(f"Running command: {' '.join(command)}")

        process = command_execution_func(command, can_be_interactive)

        if can_be_interactive and stdin_data:
            process.stdin.write(stdin_data.encode())
            process.stdin.close()

        out_thread = threading.Thread(target=reader, args=[process.stdout, sys.stdout.write])
        err_thread = threading.Thread(target=reader, args=[process.stderr, sys.stderr.write])
        out_thread.start()
        err_thread.start()
        out_thread.join()
        err_thread.join()

        return_code = process.wait()
    except (subprocess.SubprocessError, OSError) as e:
        log.debug(f"Error executing command: {e}")
        return None
    except KeyboardInterrupt:
        log.debug("Interrupted by user")
        return None
    finally:
        if process and can_be_interactive and stdin_data:
            process.stdin.close()
        return return_code


def resolve_path(path):
    return os.path.abspath(os.path.expanduser(path))



import logging
from hbox import config
from hbox.logger import set_log_level


def setup_logger():
    cfg = config.load_config()
    set_log_level(logging.DEBUG if cfg.debug else logging.INFO)


def run():
    setup_logger()

    import argparse
    from hbox.commands import (
        add_package,
        list_packages,
        remove_package,
        run_package,
        show_info,
        set_package_version,
        show_version
    )

    parser = argparse.ArgumentParser(description="CLI tool that leverages container technology to manage packages.")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    subparsers.add_parser("info", help="Print debug information.")

    subparsers.add_parser("version", help="Show current hbox version.")

    parser_list = subparsers.add_parser("list", help="List all installed packages and their versions.")
    parser_list.add_argument("name", type=str, nargs='?', default=None, help="Show all versions for a given package.")

    parser_add = subparsers.add_parser("add", help="Add a specific version of a package", aliases=['install'])
    parser_add.add_argument("name", type=str, help="Name of the package to install")
    parser_add.add_argument("version", nargs="?", type=str, default="latest",
                            help="Version of the package (default: latest)")
    parser_add.add_argument("-d", "--set-default", action="store_true",
                            help="Set the added version as the current version")

    parser_remove = subparsers.add_parser('remove', help='Remove a package.', aliases=['uninstall'])
    parser_remove.add_argument('name', help='The name of the package to remove.')
    parser_remove.add_argument("version", nargs="?", type=str, default=None,
                               help="Version of the package (default: latest)")

    parser_run = subparsers.add_parser("run", help="Run the package.")
    parser_run.add_argument("name", type=str, help="Name of the package to run")
    parser_run.add_argument("subcommand", nargs=argparse.REMAINDER, help="Arguments to pass to the package")

    parser_set = subparsers.add_parser("use", help="Set current version of a package.", aliases=["set"])
    parser_set.add_argument("name", type=str, help="Name of the package to set the version of.")
    parser_set.add_argument("version", type=str, help="New version to set as current.")

    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
    elif args.command == "info":
        show_info()
    elif args.command == "version":
        show_version()
    elif args.command == "list":
        list_packages(args.name)
    elif args.command in ["add", "install"]:
        add_package(args.name, args.version, args.set_default)
    elif args.command in ["remove", "uninstall"]:
        remove_package(args.name, args.version)
    elif args.command in ["use", "set"]:
        set_package_version(args.name, args.version)
    elif args.command == "run":
        run_package(args.name, args.subcommand)


if __name__ == "__main__":
    run()



import importlib.metadata
import os
import json
from typing import Dict, Optional, List

from pydantic import BaseModel, Field

from hbox.utils import resolve_path

lib_name = "hbox"
base_dir = resolve_path(os.getenv("HBOX_DIR", os.path.expanduser(f"~/.{lib_name}")))
config_path = resolve_path(os.path.join(base_dir, "config.json"))
versions_path = resolve_path(os.path.join(base_dir, "versions.json"))
shims_path = resolve_path(os.path.join(base_dir, "shims"))


class VolumeConfig(BaseModel):
    source: str
    target: str


class PackageConfig(BaseModel):
    image: str
    volumes: Optional[List[VolumeConfig]] = Field(default_factory=list)


class Config(BaseModel):
    debug: Optional[bool] = False
    packages: Dict[str, PackageConfig] = Field(default_factory=dict)


class Package(BaseModel):
    name: str
    versions: List[str]
    current: Optional[str] = None


class Packages(BaseModel):
    packages: List[Package] = Field(default_factory=list)


def load_config() -> Config:
    if not os.path.isfile(config_path):
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)
        default_config = Config()
        with open(config_path, "w") as f:
            f.write(default_config.model_dump_json(indent=2))
        return default_config
    with open(config_path, "r") as f:
        return Config(**json.load(f))


def save_config(data: Config):
    with open(config_path, "w") as f:
        f.write(data.model_dump_json(indent=2))


def load_versions() -> Packages:
    if not os.path.isfile(versions_path):
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)
        default_packages = Packages()
        with open(versions_path, "w") as f:
            f.write(default_packages.model_dump_json(indent=2))
        return default_packages
    with open(versions_path, "r") as f:
        return Packages(**json.load(f))


def save_versions(data: Packages):
    with open(versions_path, "w") as f:
        f.write(data.model_dump_json(indent=2))


def get_library_version() -> str:
    try:
        version = importlib.metadata.version(lib_name)
        return version
    except importlib.metadata.PackageNotFoundError:
        return "<unknown>"



import os
import platform
import subprocess
from typing import Optional

import hbox.config as config
from hbox.logger import get_logger
from hbox.utils import execute_command, resolve_path

log = get_logger(__name__)


def add_shim(name: str):
    shims_file_path = resolve_path(os.path.join(config.shims_path, name))
    if not os.path.exists(shims_file_path):
        os.makedirs(os.path.dirname(shims_file_path), exist_ok=True)
        with open(shims_file_path, 'w') as shim_file:
            shim_file.write('#!/bin/sh\n')
            shim_file.write('hbox run ' + name + ' "$@"\n')

        # Set the permissions of the new shim script to be executable (0o755 gives rwx for user and rx for group/others)
        os.chmod(shims_file_path, 0o755)


def remove_shim(name: str):
    shims_file_path = resolve_path(os.path.join(config.shims_path, name))
    if os.path.exists(shims_file_path):
        os.remove(shims_file_path)


def get_container_image_url(name: str, version: str, set_as_default: bool):
    cfg = config.load_config()
    versions_cfg = config.load_versions()
    found = False
    for current_package in versions_cfg.packages:
        if current_package.name == name:
            if version in current_package.versions:
                log.info(f"'{name}' version {version} already exists.")
                return None, versions_cfg
            else:
                current_package.versions.append(version)
                if set_as_default:
                    log.info(f"'{name}' version {version} set as default.")
                    current_package.current = version
                found = True
            break
    if not found:
        package = config.Package(name=name, versions=[version], current=version)
        versions_cfg.packages.append(package)

    if name in cfg.packages:
        image = cfg.packages[name].image
    else:
        image = name

    return f"{image}:{version}", versions_cfg


def add_package(name, version: Optional[str] = "latest", set_as_default: Optional[bool] = False):
    image_url, versions_cfg = get_container_image_url(name, version, set_as_default)
    if image_url:
        full_command = ["docker", "pull", image_url]
        exit_code = execute_command(full_command)
        if exit_code and exit_code != 0:
            log.error(f"Failed to add package '{name}' at version {version}.")
        else:
            config.save_versions(versions_cfg)
            log.info(f"Added '{name}' version {version}.")
            add_shim(name)


def remove_package(name: str, version: Optional[str] = None):
    versions_cfg = config.load_versions()
    for current_package in versions_cfg.packages:
        if current_package.name == name:
            if version:
                if version in current_package.versions:
                    # Check if version to be removed is the current one and if there are multiple versions
                    if version == current_package.current and len(current_package.versions) > 1:
                        log.error(
                            f"Cannot remove the current active version '{version}' of '{name}'.")
                        return
                    current_package.versions.remove(version)
                    log.info(f"Removed version '{version}' of '{name}'.")
                else:
                    log.error(f"Version '{version}' of '{name}' does not exist.")

            if not version or not current_package.versions:
                versions_cfg.packages.remove(current_package)
                remove_shim(name)
                log.info(f"Removed package '{name}'.")
            break
    else:
        log.error(f"Package '{name}' does not exist.")

    config.save_versions(versions_cfg)


def run_package(name: str, command: list):
    cfg = config.load_config()
    versions_cfg = config.load_versions()

    image = name
    version = "latest"

    volumes_command = []
    if name in cfg.packages:
        package = cfg.packages[name]
        image = package.image
        for volume in package.volumes:
            source = resolve_path(volume.source)
            target = volume.target
            if os.path.exists(source):
                log.debug(f"Mounting volume {source} to {target}")
                volumes_command.extend(["-v", f"{source}:{target}"])
            else:
                log.debug(f"Volume {source} not found. Skipping.")
    else:
        log.debug(f"No configuration found for package '{name}'. Using {name} as the image name.")

    for current_package in versions_cfg.packages:
        if current_package.name == name:
            version = current_package.current

    full_image_url = f"{image}:{version}"

    full_command = ["docker", "run", "--rm", full_image_url] + volumes_command + command
    try:
        execute_command(full_command, can_be_interactive=True)
    except subprocess.CalledProcessError as e:
        log.debug(f"Failed to run command {' '.join(full_command)}")


def set_package_version(name: str, version: str):
    versions_cfg = config.load_versions()
    for current_package in versions_cfg.packages:
        if current_package.name == name:
            if version in current_package.versions:
                current_package.current = version
                config.save_versions(versions_cfg)
                log.info(f"'{name}' set to version {version}")
            else:
                log.error(f"'{name}' version {version} not found. Add the version first via 'add' command.")


def show_info():
    log.info("OS:")
    os_info = ' '.join(platform.uname())
    log.info(os_info)

    log.info("\nHBOX VERSION:")
    log.info(config.get_library_version())

    log.info("\nHBOX ENVIRONMENT VARIABLES:")
    log.info(f"HBOX_DIR={config.base_dir}")


def show_version():
    log.info(config.get_library_version())


def print_package(package):
    log.info(f"- {package.name}:")
    sorted_versions = sorted(package.versions)
    for version in sorted_versions:
        msg = f"  - {version}"
        if version == package.current:
            msg += " ✔"
        log.info(msg)


def list_packages(name: Optional[str] = None):
    versions_cfg = config.load_versions()
    if name:
        found = False
        for package in versions_cfg.packages:
            if package.name == name:
                print_package(package)
                found = True
                break
        if not found:
            log.error(f"Package '{name}' was not found. Add the package first via 'add' command.")
    else:
        for package in sorted(versions_cfg.packages, key=lambda pkg: pkg.name):
            print_package(package)



import logging


class CustomFormatter(logging.Formatter):
    def format(self, record):
        if record.levelno != logging.INFO:
            self._style._fmt = "[%(levelname)s] %(message)s"
        else:
            self._style._fmt = "%(message)s"
        return super().format(record)


def get_logger(name: str):
    global log_level
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    formatter = CustomFormatter()
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger


def set_log_level(level):
    global log_level
    log_level = level


log_level = logging.INFO
